-------------------------------------------------------------------------------
-- Title      : controlUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Control unit of a sequential FIR filter.
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2018-04-11  1.0      jnbazin Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity controlUnit is

  port (
    I_clock               : in  std_logic;  -- global clock
    I_reset               : in  std_logic;  -- asynchronous global reset
    I_inputSampleValid    : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_processingDone      : in  std_logic;
    O_loadShift1           : out std_logic;  -- filtered sample
    O_loadShift2           : out std_logic;  -- filtered sample
    O_loadShift3           : out std_logic;  -- filtered sample
    O_initAddress         : out std_logic;  -- Control signal to initialize register read address
    O_incrAddress         : out std_logic;  -- Control signal to increment register read address
    O_initSum             : out std_logic;  -- Control signal to initialize the MAC register
    O_loadSum             : out std_logic;  -- Control signal to load the MAC register;
    O_loadY               : out std_logic;  -- Control signal to load Y register
    O_coefSelect          : out std_logic_vector(1 downto 0);
    O_FilteredSampleValid : out std_logic  -- Data valid signal for filtered sample
    );

end entity controlUnit;
architecture archi_operativeUnit of controlUnit is


  type T_state is (WAIT_SAMPLE, STORE, PROCESSING_LOOP_1, PROCESSING_LOOP_2_1, PROCESSING_LOOP_2_2, PROCESSING_LOOP_3, OUTPUT_1, OUTPUT_2, OUTPUT_3, ADDRESS_RESET, WAIT_END_SAMPLE);  -- state list
  signal SR_presentState : T_state;
  signal SR_futurState   : T_state;

begin

  process (I_clock,I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_presentState <= WAIT_SAMPLE;
    elsif rising_edge(I_clock) then     -- rising clock edge
      SR_presentState <= SR_futurState;
    end if;
  end process;

  process (SR_presentState, I_inputSampleValid, I_processingDone) is
  begin
    case SR_presentState is

      when WAIT_SAMPLE =>
        if I_inputSampleValid = '1' then
          SR_futurState <= STORE;
        else 
          SR_futurState <= WAIT_SAMPLE;
        end if;

      when STORE =>
          SR_futurState <= PROCESSING_LOOP_1;

      when PROCESSING_LOOP_1 =>
      O_coefSelect <= "00";
        if I_processingDone = '1' then
          SR_futurState <= OUTPUT_1;
        else 
          SR_futurState <= PROCESSING_LOOP_1;
        end if;
      
      when OUTPUT_1 =>
        SR_futurState <= PROCESSING_LOOP_2_1;

      when PROCESSING_LOOP_2_1 =>
      O_coefSelect <= "01";
        if I_processingDone = '1' then
          SR_futurState <= ADDRESS_RESET;
        else 
          SR_futurState <= PROCESSING_LOOP_2_1;
        end if;

      when ADDRESS_RESET =>
        SR_futurState <= PROCESSING_LOOP_2_2;

      when PROCESSING_LOOP_2_2 =>
      O_coefSelect <= "10";
        if I_processingDone = '1' then
          SR_futurState <= OUTPUT_2;
        else 
          SR_futurState <= PROCESSING_LOOP_2_2;
        end if;

      when OUTPUT_2 =>
        SR_futurState <= PROCESSING_LOOP_3;

      when PROCESSING_LOOP_3 =>
      O_coefSelect <= "11";
        if I_processingDone = '1' then
          SR_futurState <= OUTPUT_3;
        else 
          SR_futurState <= PROCESSING_LOOP_3;
        end if;

      when OUTPUT_3 =>
        SR_futurState <= WAIT_END_SAMPLE;

      when WAIT_END_SAMPLE =>
        if I_inputSampleValid = '0' then
          SR_futurState <= WAIT_SAMPLE;
        else 
          SR_futurState <= WAIT_END_SAMPLE;
        end if;

      when others => null;
    end case;
  end process;

  O_loadShift1           <= '1' when SR_presentState = Store else '0';
  O_loadShift2           <= '1' when SR_presentState = OUTPUT_1 else '0';
  O_loadShift3           <= '1' when SR_presentState = OUTPUT_2 else '0';
  O_initAddress         <= '1' when SR_presentState = Store or SR_presentState = OUTPUT_1 or SR_presentState = OUTPUT_2 or SR_presentState = ADDRESS_RESET else '0';
  O_incrAddress         <= '1' when SR_presentState = PROCESSING_LOOP_1 or SR_presentState = PROCESSING_LOOP_2_1 or SR_presentState = PROCESSING_LOOP_2_2 or SR_presentState = PROCESSING_LOOP_3 else '0';
  O_initSum             <= '1' when SR_presentState = Store or SR_presentState = OUTPUT_1 or SR_presentState = OUTPUT_2 else '0';
  O_loadSum             <= '1' when SR_presentState = PROCESSING_LOOP_1 or SR_presentState = PROCESSING_LOOP_2_1 or SR_presentState = PROCESSING_LOOP_2_2 or SR_presentState = PROCESSING_LOOP_3 else '0';
  O_loadY               <= '1' when SR_presentState = OUTPUT_3 else '0';
  O_FilteredSampleValid <= '1' when SR_presentState = WAIT_END_SAMPLE else '0';





end architecture archi_operativeUnit;
