-------------------------------------------------------------------------------
-- Title      : operativeUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Operative unit of a sequential FIR filter. Including shift
-- register for samples, registers for coefficients, a MAC and a register to
-- store the result
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2019-02-13  1.1      marzel  Update to provide a 16-tap filter and improve
--                              the user experience ;)
-- 2018-04-11  1.0      jnbazin Created
-- 2018-04-18  1.0      marzel  Modification of SR_Y assignment to a round
--                              instead of a trunc
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity operativeUnit is

  port (
    I_clock          : in  std_logic;   -- global clock
    I_reset          : in  std_logic;   -- asynchronous global reset
    I_inputSample    : in  std_logic_vector(10 downto 0);  -- 8 bit input sample
    I_loadShift1           : in std_logic;  -- filtered sample
    I_loadShift2           : in std_logic;  -- filtered sample
    I_loadShift3           : in std_logic;  -- filtered sample
    I_initAddress         : in std_logic;  -- Control signal to initialize register read address
    I_incrAddress         : in std_logic;  -- Control signal to increment register read address
    I_initSum             : in std_logic;  -- Control signal to initialize the MAC register
    I_loadSum             : in std_logic;  -- Control signal to load the MAC register;
    I_loadY               : in std_logic;  -- Control signal to load Y register
    I_coefSelect          : in std_logic_vector(1 downto 0);
    O_processingDone : out std_logic;   -- Indicate that processing is done
    O_Y              : out std_logic_vector(10 downto 0)   -- filtered sample
    );

end entity operativeUnit;

architecture arch_operativeUnit of operativeUnit is
  signal SC_smpl  : signed(10 downto 0);
  signal SC_coef  : signed(10 downto 0);
  signal SC_result     : signed(28 downto 0);  -- result of the accumulation addition
  signal SR_Y             : signed(10 downto 0);  -- filtered sample storage register
  signal SR_readAddress   : integer range 0 to 94;  -- register files read address

  signal SR_readAddress2_1   : integer range 0 to 2;  -- register files read address
  signal SR_readAddress2_2   : integer range 0 to 1;  -- register files read address
  signal SR_readAddress3   : integer range 0 to 10;  -- register files read address
  

  type registerFile1 is array(0 to 94) of signed(10 downto 0);
  signal sampleReg1        : registerFile1;
  type registerFile2 is array(0 to 2) of signed(10 downto 0);
  signal sampleReg2        : registerFile2;
  type registerFile3 is array(0 to 10) of signed(10 downto 0);
  signal sampleReg3        : registerFile3;
  signal maxSum            : integer;

  signal SR_coefRegister1        : registerFile1;
  signal SR_coefRegister2_1        : registerFile2;
  type registerFile2_2 is array(0 to 1) of signed(10 downto 0);
  signal SR_coefRegister2_2        : registerFile2_2;
  signal SR_coefRegister3        : registerFile3;

  component mac is
    Port ( clk        : in  STD_LOGIC;
           rst        : in  STD_LOGIC;
           sum_rst    : in std_logic;
           enable: in std_logic;
           smpl          : in  signed(10 downto 0); -- 11-bit input a
           coef          : in  signed(10 downto 0); -- 11-bit input b
           acc_out    : inout signed(28 downto 0) -- 29-bit acc output
           );
  end component mac;

  component mux4to1 is
    generic (
        DATA_WIDTH : integer := 11 -- Default width of the input data (can be changed)
    );
    Port ( 
        sel        : in  STD_LOGIC_VECTOR(1 downto 0); -- 2-bit select input to choose between 4 entries
        a          : in  signed(DATA_WIDTH-1 downto 0); -- 1st input (configurable width)
        b          : in  signed(DATA_WIDTH-1 downto 0); -- 2nd input (configurable width)
        c          : in  signed(DATA_WIDTH-1 downto 0); -- 3rd input (configurable width)
        d          : in  signed(DATA_WIDTH-1 downto 0); -- 4th input (configurable width)
        mux_out    : out signed(DATA_WIDTH-1 downto 0) -- Output with configurable width
    );
end component mux4to1;

begin

  mac1 : mac
    port map (
      clk        => I_clock,
      rst        => I_reset,
      sum_rst    => I_initSum,
      enable     => I_loadSum,
      smpl       => SC_smpl,
      coef       => SC_coef,
      acc_out    => SC_result
    );

--  mux_coef : mux4to1 
--    port map(
--      sel        => I_coefSelect,
--      a          => SR_coefRegister1(SR_readAddress),
--      b          => SR_coefRegister2_1(SR_readAddress),
--      c          => SR_coefRegister2_2(SR_readAddress),
--      d          => SR_coefRegister3(SR_readAddress),
--      mux_out    => SC_coef
--      );
      
      
    SR_readAddress2_1 <=   SR_readAddress when(I_coefSelect = "01") else 0;
    SR_readAddress2_2 <=   SR_readAddress when(I_coefSelect = "10") else 0;
    SR_readAddress3 <=   SR_readAddress when(I_coefSelect = "11") else 0;
      
    SC_coef <=  SR_coefRegister1(SR_readAddress)   when(I_coefSelect = "00")else
                SR_coefRegister2_1(SR_readAddress2_1) when(I_coefSelect = "01")else
                SR_coefRegister2_2(SR_readAddress2_2) when(I_coefSelect = "10")else
                SR_coefRegister3(SR_readAddress3);
    
--  mux_smpl : mux4to1 
--    port map(
--      sel        => I_coefSelect,
--      a          => sampleReg1(SR_readAddress),
--      b          => sampleReg2(SR_readAddress2_1),
--      c          => sampleReg3(SR_readAddress2_2),
--      d          => sampleReg3(SR_readAddress3),
--      mux_out    => SC_smpl
--      );
      
      SC_smpl <=  sampleReg1(SR_readAddress)   when(I_coefSelect = "00")else
                sampleReg2(SR_readAddress2_1) when(I_coefSelect = "01")else
                sampleReg3(SR_readAddress3);

  shift : process (I_clock, I_reset) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      sampleReg1 <= (others => (others => '0'));
      sampleReg2 <= (others => (others => '0'));
      sampleReg3 <= (others => (others => '0'));
    elsif rising_edge(I_clock) then
      if I_loadShift1 = '1' then

        for i in 0 to 93 loop 

        sampleReg1(i+1) <= sampleReg1(i);

        end loop;

        sampleReg1(0) <= Signed(I_inputSample);
      
      elsif I_loadShift2 = '1' then

        for i in 0 to 1 loop 
  
        sampleReg2(i+1) <= sampleReg2(i);

       end loop;

       sampleReg2(0) <= signed(SC_result(20 downto 10));

      elsif I_loadShift3 = '1' then

        for i in 0 to 9 loop 

        sampleReg3(i+1) <= sampleReg3(i);

        end loop;

        sampleReg3(0) <= signed(SC_result(19 downto 9));

      end if;

    end if;
  end process shift;

  incr_address : process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_readAddress <= 0;
    
    elsif rising_edge(I_clock) then
      if I_initAddress = '1' then
        SR_readAddress <= 0; 
      elsif I_incrAddress = '1' then
        SR_readAddress <= SR_readAddress+1;
      end if;

    end if;
  end process incr_address;

  maxSumSel : process (I_coefSelect) is
  begin 
    case I_coefSelect is
      when "00" =>
        maxSum <= 93;
      when "01" =>
        maxSum <= 1;
      when "10" =>
        maxSum <= 0;
      when "11" =>
        maxSum <= 9;
      when others => 
        maxSum <= 0;
    end case;
  end process maxSumSel;
 
  O_processingDone <= '1' when SR_readAddress = maxSum else '0' ; 

  -- SC_multOperand1 <= SR_shiftRegister(SR_readAddress) ;   -- 8 bits
  -- SC_multOperand2 <= SR_coefRegister(SR_readAddress) ;    -- 8 bits
  -- SC_MultResult   <= SC_multOperand1 * SC_multOperand2 ;  -- 16 bits
  -- SC_addResult    <= resize(SC_MultResult, SC_addResult'length) + SR_sum;

  -- sum_acc : process (I_clock, I_reset) is
  -- begin
  --   if I_reset = '1' then               -- asynchronous reset (active high)
  --     SR_sum <= (others => '0');
  --   elsif rising_edge(I_clock) then
  --     if I_initSum = '1' then
  --       SR_sum <= (others => '0');
  --     elsif I_loadSum = '1' then
  --       SR_sum <= SC_addResult;
  --     end if;

  --   end if;
  -- end process sum_acc;

  store_result : process (I_clock, I_reset) is
  begin
      if I_reset = '1' then               -- asynchronous reset (active high)
        SR_Y <= (others => '0');
      elsif rising_edge(I_clock) then
        if I_loadY = '1' then
          SR_Y <= SC_result(20 downto 10); --TODO: changer la sortie
        end if;
  
      end if;
  end process store_result;

  O_Y <= std_logic_vector(SR_Y);


-- Low-pass filter provided with octave (or Matlab ;)) command
--fir1(15, .001)/sqrt(sum(fir1(15, .001).^2))*2^6
SR_coefRegister1 <= (to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-2, 11),
to_signed(-2, 11),
to_signed(-2, 11),
to_signed(-3, 11),
to_signed(-3, 11),
to_signed(-3, 11),
to_signed(-4, 11),
to_signed(-4, 11),
to_signed(-5, 11),
to_signed(-5, 11),
to_signed(-6, 11),
to_signed(-6, 11),
to_signed(-7, 11),
to_signed(-7, 11),
to_signed(-8, 11),
to_signed(-8, 11),
to_signed(-9, 11),
to_signed(-10, 11),
to_signed(-10, 11),
to_signed(-11, 11),
to_signed(-11, 11),
to_signed(-12, 11),
to_signed(-13, 11),
to_signed(-13, 11),
to_signed(-14, 11),
to_signed(-14, 11),
to_signed(-15, 11),
to_signed(-15, 11),
to_signed(-16, 11),
to_signed(-16, 11),
to_signed(-17, 11),
to_signed(-17, 11),
to_signed(-18, 11),
to_signed(-18, 11),
to_signed(-18, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(1004, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-19, 11),
to_signed(-18, 11),
to_signed(-18, 11),
to_signed(-18, 11),
to_signed(-17, 11),
to_signed(-17, 11),
to_signed(-16, 11),
to_signed(-16, 11),
to_signed(-15, 11),
to_signed(-15, 11),
to_signed(-14, 11),
to_signed(-14, 11),
to_signed(-13, 11),
to_signed(-13, 11),
to_signed(-12, 11),
to_signed(-11, 11),
to_signed(-11, 11),
to_signed(-10, 11),
to_signed(-10, 11),
to_signed(-9, 11),
to_signed(-8, 11),
to_signed(-8, 11),
to_signed(-7, 11),
to_signed(-7, 11),
to_signed(-6, 11),
to_signed(-6, 11),
to_signed(-5, 11),
to_signed(-5, 11),
to_signed(-4, 11),
to_signed(-4, 11),
to_signed(-3, 11),
to_signed(-3, 11),
to_signed(-3, 11),
to_signed(-2, 11),
to_signed(-2, 11),
to_signed(-2, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11),
to_signed(-1, 11));


SR_coefRegister2_1 <= (to_signed(480, 11),
to_signed(-777, 11),
to_signed(480, 11));

SR_coefRegister2_2 <= (to_signed(777, 11),
to_signed(-449, 11));

SR_coefRegister3 <= (to_signed(-119, 11),
to_signed(122, 11),
to_signed(149, 11),
to_signed(191, 11),
to_signed(226, 11),
to_signed(239, 11),
to_signed(226, 11),
to_signed(191, 11),
to_signed(149, 11),
to_signed(122, 11),
to_signed(-119, 11));



end architecture arch_operativeUnit;
